#pragma once

#include <string>
#include <cstdint>
#include <map>

namespace Tencent {
namespace WxWork {

class SpecSDKImpl;
class SpecCallbackSDKImpl;
class SpecCallbackSDK;

/*
 * @description                 开启调试模式,进程级别开关
 * @param debug_token:          调试凭证,从管理端拿到
 * @param access_token:         应用access_token
 * @return:                     是否开启成功
 * */
bool SpecOpenDebugMode(
        const std::string &debug_token,
        const std::string &access_token);

/*
 * @description 打印日志
 * */
enum LogLevel {
    eLogLevelError = 1,
    eLogLevelInfo = 2,
    eLogLevelDebug = 3,
};
void SpecLog(
        LogLevel level, const char *filename, int line, const char * fmt, ...);
#define WWSPEC_LOG(LEVEL, FORMAT, ...) \
    Tencent::WxWork::SpecLog(LEVEL, __FILE__, __LINE__, FORMAT, ##__VA_ARGS__)
#define WWSPEC_LOG_ERR(FORMAT, ...) \
    WWSPEC_LOG(Tencent::WxWork::eLogLevelError, FORMAT, ##__VA_ARGS__)
#define WWSPEC_LOG_INFO(FORMAT, ...) \
    WWSPEC_LOG(Tencent::WxWork::eLogLevelInfo, FORMAT, ##__VA_ARGS__)
#define WWSPEC_LOG_DEBUG(FORMAT, ...) \
    WWSPEC_LOG(Tencent::WxWork::eLogLevelDebug, FORMAT, ##__VA_ARGS__)

/*
 * @description:    一键生成程序内独立的notify id
 * @return:         notify id
 * */
std::string GenerateNotifyId();

class SpecSDK {
public:
    /*
     * @param corpid:     企业corpid
     * @param agentid:    应用id
     * @param ability_id: 能力ID
     * @param job_info:   job_info
     * @param 
     * */
    SpecSDK(
            const std::string & corpid, 
            uint32_t agentid, 
            const std::string & ability_id = "",
            const std::string & job_info = "");

    /*
     * @description         使用callback的请求来初始化
     * @param callback_sdk: 要求IsOk()==true
     * */
    SpecSDK(const SpecCallbackSDK &callback_sdk);

    ~SpecSDK();

    /*
     * @description:    用于在专区内调用企业微信接口
     * @param api_name: 接口名
     * @param request:  json格式的请求数据
     * @param response: json格式的返回数据
     * @return:         当0时表示没有网络或请求协议层面的失败,调用方需继续检查response中的errcode字段
     * @example:        Invoke("sync_msg", R"({"cursor":"CURSOR","token":"TOKEN","limit":200})", &response);
     * */

    /*
     * 当前版本sdk支持的接口列表,每个接口的具体协议请查看企业微信文档:
     * https://developer.work.weixin.qq.com/document/path/99819
     * +--------------------------------+--------------------------------+
     * |接口名                          |描述                            |
     * |--------------------------------|--------------------------------|
     * |program_async_job_call_back     |上报异步任务结果                |
     * |sync_msg                        |获取会话记录                    |
     * |get_group_chat                  |获取内部群信息                  |
     * |get_agree_status_single         |获取单聊会话同意情况            |
     * |get_agree_status_room           |获取群聊会话同意情况            |
     * |set_hide_sensitiveinfo_config   |设置成员会话组件敏感信息隐藏配置|
     * |get_hide_sensitiveinfo_config   |获取成员会话组件敏感信息隐藏配置|
     * |search_chat                     |会话名称搜索                    |
     * |search_msg                      |会话消息搜索                    |
     * |create_rule                     |新增关键词规则                  |
     * |get_rule_list                   |获取关键词列表                  |
     * |get_rule_detail                 |获取关键词规则详情              |
     * |update_rule                     |修改关键词规则                  |
     * |delete_rule                     |删除关键词规则                  |
     * |get_hit_msg_list                |获取命中关键词规则的会话记录    |
     * |create_sentiment_task           |创建情感分析任务                |
     * |get_sentiment_result            |获取情感分析结果                |
     * |create_summary_task             |创建摘要提取任务                |
     * |get_summary_result              |获取摘要提取结果                |
     * |create_customer_tag_task        |创建标签匹配任务                |
     * |get_customer_tag_result         |获取标签任务结果                |
     * |create_recommend_dialog_task    |创建话术推荐任务                |
     * |get_recommend_dialog_result     |获取话术推荐结果                |
     * |create_private_task             |创建自定义模型任务              |
     * |get_private_task_result         |获取自定义模型结果              |
     * |(废弃)document_list             |获取知识集列表                  |
     * |create_spam_task                |会话反垃圾创建分析任务          |
     * |get_spam_result                 |会话反垃圾获取任务结果          |
     * |create_chatdata_export_job      |创建会话内容导出任务            |
     * |get_chatdata_export_job_status  |获取会话内容导出任务结果        |
     * |spec_notify_app                 |专区通知应用                    |
     * |create_program_task             |创建自定义程序任务              |
     * |get_program_task_result         |获取自定义程序结果              |
     * |knowledge_base_list             |获取企业授权给服务商的知识集列表|
     * |knowledge_base_create           |创建知识集                      |
     * |knowledge_base_detail           |获取知识集详情                  |
     * |knowledge_base_add_doc          |添加知识集內容                  |
     * |knowledge_base_remove_doc       |删除知识集內容                  |
     * |knowledge_base_modify_name      |修改知识集名称                  |
     * |knowledge_base_delete           |删除知识集                      |
     * +-----------------------------------------------------------------+
     * */
    int Invoke(const std::string &api_name, const std::string &request, std::string *response);

private:
    SpecSDKImpl * impl_;
};

class SpecCallbackSDK {
public:
    /*
     * @description:   每个请求构造一个SpecCallbackSDK示例,
     *                 SpecCallbackSDK仅持有headers和body的引用,
     *                 因此需保证headers和body的生存期比SpecCallbackSDK长
     * @param method:  请求方法GET/POST
     * @param headers: 请求header
     * @param body:    请求body
     * @example: 
     * SpecCallbackSDK sdk(method, headers, body);
     * string response;  //plain
     * if (sdk.IsOk()) {
     *   auto corpid = sdk.GetCorpId();
     *   auto agentid = sdk.GetAgentId();
     *   auto call_type = sdk.GetCallType();
     *   auto &data = sdk.GetData();
     *   //do something
     *   response = "{}";
     * } 
     * map<string, string> response_headers;
     * string body;      //encrypted
     * sdk.BuildResponseHeaderBody(content, &response_headers, &body);
     * //do response
     * */
    SpecCallbackSDK(
            const std::string &method,
            const std::map<std::string, std::string> & headers,
            const std::string &body);
    ~SpecCallbackSDK();

    /*
     * @description: 判断构造函数中传入的请求是否解析成功
     * @return:      成功与否
     * */
    bool IsOk() const;

    /*
     * @description: 获取请求的企业
     * @require:     仅当IsOk() == true可调用
     * @return:      corpid
     * */
    const std::string &GetCorpId() const;

    /*
     * @description: 获取请求的应用
     * @require:     仅当IsOk() == true可调用
     * @return:      agentid
     * */
    uint32_t GetAgentId() const;

    /*
     * @description: 获取请求的类型
     * @require:     仅当IsOk() == true可调用
     * @return:      1 - 来自[应用调用专区]的请求
     *               2 - 来自企业微信的回调事件
     * */
    uint32_t GetCallType() const;
    
    /*
     * @description: 获取请求数据
     * @require:     仅当IsOk() == true可调用
     * @return:      请求数据,根据call_type可能是:
     *               - 企业微信回调事件
     *               - [应用调用专区]接口中的request_data
     * */
    const std::string &GetData() const;

    /*
     * @description: 是否异步请求
     * @require:     仅当IsOk() == true可调用
     * @return:      
     * */
    bool GetIsAsync() const;

    /*
     * @description: 获取请求的job_info,
     * @require:     仅当IsOk() == true可调用
     * @return:      job_info,无需理解内容,
     *               在同一个请求上下文中使用SpecSDK的时候传入
     * */
    const std::string &GetJobInfo() const;

    /*
     * @description: 获取请求的ability_id,[应用调用专区]接口时指定
     * @require:     仅当IsOk() == true可调用
     * @return:      ability_id
     * */
    const std::string &GetAbilityId() const;

    /*
     * @description: 获取请求的notify_id,用于[应用同步调用专区程序]接口
     * @require:     仅当IsOk() == true可调用
     * @return:      ability_id
     * */
    const std::string &GetNotifyId() const;

    /*
     * @description:    对返回包计算签名&加密
     * @param response: 待加密的回包明文.如果IsOk()==false,传入空串即可
     * @param headers:  回包的header
     * @param body:     密文body
     * */
    void BuildResponseHeaderBody(
            const std::string & response,
            std::map<std::string, std::string> *headers,
            std::string *body);
private:
    SpecCallbackSDKImpl * impl_;
};

}
}
