#include "cppspecdemo_db.h"
#include "cpp_sdk.h"
#include <mutex>
#include <map>
#include <queue>
#include <thread>
#include <stdio.h>
#include <sys/stat.h>
#include <error.h>
#include <dirent.h>
#include <unistd.h>
using namespace std;

#define LogErr(FORMAT, ...) WWSPEC_LOG_ERR(FORMAT, ##__VA_ARGS__)
#define LogInfo(FORMAT, ...) WWSPEC_LOG_INFO(FORMAT, ##__VA_ARGS__)

class DBImpl_File: public DB {
public:
    DBImpl_File();
    virtual ~DBImpl_File();
    virtual int SetCallbackEvent(
            const std::string &corpid,
            uint32_t agentid,
            const std::string &notify_id,
            const std::string &event_data);
    virtual int GetCallbackEvent(
            const std::string &corpid,
            uint32_t agentid,
            const std::string &notify_id,
            std::string *event_data);
private:
    std::string GetFileName(const std::string &corpid,
            uint32_t agentid, 
            const std::string &notify_id) {
        return callback_notify_dir_ 
            + "/" + corpid 
            + "_" + to_string(agentid) 
            + "_" + notify_id;
    }
    
    static void ClearTimeoutNotifyIdThread();

    static const std::string &callback_notify_dir_;
};
const std::string &DBImpl_File::callback_notify_dir_ = "/mnt/data/callback_notify";

DBImpl_File::DBImpl_File() {
    int ret = mkdir(callback_notify_dir_.c_str(), 
            S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
    if (ret != 0 && errno != EEXIST) {
        LogErr("mkdir fail, abort");
        abort();
    }

    new std::thread(&DBImpl_File::ClearTimeoutNotifyIdThread);
}

DBImpl_File::~DBImpl_File() {
}

void DBImpl_File::ClearTimeoutNotifyIdThread() {
    while (true) {
        LogInfo("scan and clear timeout notify_id");
        time_t now = time(NULL);
        DIR * dir = opendir(callback_notify_dir_.c_str());
        if (dir == NULL) {
            LogErr("open dir fail");
        } else {
            struct dirent *item;
            while ((item = readdir(dir)) != NULL) {
                if (item->d_name[0] == '.') continue;
                string filename = callback_notify_dir_ + "/" + item->d_name;
                struct stat st;
                if (stat(filename.c_str(), &st) != 0) {
                    LogErr("stat file fail: %s", filename.c_str());
                    continue;
                }
                if (S_ISREG(st.st_mode)) {
                    //删除超过20分钟后的文件
                    if (st.st_mtime + 1200 < now) {
                        LogInfo("remove notify_id: %s", filename.c_str());
                        remove(filename.c_str());
                    }
                }
            }
            closedir(dir);
        }
        sleep(600);
    }
}

int DBImpl_File::SetCallbackEvent(
        const std::string &corpid,
        uint32_t agentid,
        const std::string &notify_id,
        const std::string &event_data) {
    string file_name = GetFileName(corpid, agentid, notify_id);
    FILE * fp = fopen(file_name.c_str(), "w+");
    if (fp == NULL) {
        LogErr("open file fail");
        return -2;
    }
    int ret = fwrite(event_data.c_str(), 1, event_data.size(), fp);
    if (ret < 0) {
        LogErr("write file fail");
        fclose(fp);
        return -2;
    }
    fclose(fp);
    return 0;
}
int DBImpl_File::GetCallbackEvent(
        const std::string &corpid,
        uint32_t agentid,
        const std::string &notify_id,
        std::string *event_data) {
    string file_name = GetFileName(corpid, agentid, notify_id);
    FILE * fp = fopen(file_name.c_str(), "r");
    if (fp == NULL) {
        if (errno == ENOENT) {
            LogErr("not found notify_id");
            return 1;
        }
        LogErr("open file fail");
        return -2;
    }
    char buf[4096];
    int ret;
    event_data->clear();
    while ((ret = fread(buf, 1, sizeof(buf), fp)) > 0) {
        event_data->append(buf, ret);
    }
    fclose(fp);
    return 0;
}

//begin DBImpl_LocalMem
class DBImpl_LocalMem: public DB {
public:
    DBImpl_LocalMem();
    virtual ~DBImpl_LocalMem();
    virtual int SetCallbackEvent(
            const std::string &corpid,
            uint32_t agentid,
            const std::string &notify_id,
            const std::string &event_data);

    virtual int GetCallbackEvent(
            const std::string &corpid,
            uint32_t agentid,
            const std::string &notify_id,
            std::string *event_data);
private:
    void DropExpireEventData();
private:
    struct EventDataEntry {
        std::string corpid;
        uint32_t agentid;
        std::string notify_id;
        std::string event_data;
        time_t expire_timestamp;
        EventDataEntry(
                const std::string corpid,
                uint32_t agentid,
                const std::string &notify_id,
                const std::string &event_data)
            : corpid(corpid),
            agentid(agentid),
            notify_id(notify_id),
            event_data(event_data), 
            expire_timestamp(time(NULL) + 600) {}
        static string Key(
                const std::string &corpid,
                uint32_t agentid,
                const std::string &notify_id) {
            return corpid + 
                "_" + to_string(agentid) + 
                "_" + notify_id;
        }
        string Key() const {
            return Key(corpid, agentid, notify_id);
        }
    };
    //notify_id的有效期是10分钟
    //简单地过期淘汰掉
    std::mutex m_;
    map<string, EventDataEntry *> event_data_entry_;
    queue<EventDataEntry *> event_data_expire_queue_;
};

DBImpl_LocalMem::DBImpl_LocalMem() {}
DBImpl_LocalMem::~DBImpl_LocalMem() {}

void DBImpl_LocalMem::DropExpireEventData() {
    auto &q = event_data_expire_queue_;
    while (!q.empty() && q.front()->expire_timestamp < time(NULL)) {
        auto entry = q.front(); 
        q.pop();
        event_data_entry_.erase(entry->Key());
        LogInfo("drop expired event_data: %s %s",
                entry->notify_id.c_str(), entry->event_data.c_str());
        delete entry;
    }
}

int DBImpl_LocalMem::SetCallbackEvent(
        const std::string &corpid,
        uint32_t agentid,
        const std::string &notify_id,
        const std::string &event_data) {
    auto entry = new EventDataEntry(corpid, agentid, notify_id, event_data);
    std::lock_guard<std::mutex> l(m_);
    DropExpireEventData();
    event_data_entry_[entry->Key()] = entry;
    event_data_expire_queue_.push(entry);
    LogInfo("set event_data: %s -> %s",
            notify_id.c_str(), event_data.c_str());
    return 0;
}

int DBImpl_LocalMem::GetCallbackEvent(
        const std::string &corpid,
        uint32_t agentid,
        const std::string &notify_id,
        std::string *event_data) {
    std::lock_guard<std::mutex> l(m_);
    DropExpireEventData();
    string sKey = EventDataEntry::Key(corpid, agentid, notify_id);
    auto iter = event_data_entry_.find(sKey);
    if (iter == event_data_entry_.end()) {
        LogInfo("not found notify_id: %s", sKey.c_str());
        return 1;
    }
    *event_data = iter->second->event_data;
    LogInfo("got event_data: %s -> %s",
            sKey.c_str(), event_data->c_str());
    return 0;
}

//end DBImpl_LocalMem

DB::DB() {}
DB::~DB() {}

DB * DB::Instance() {
    static DB *db = new DBImpl_File();
    //static DB *db = new DBImpl_LocalMem();
    return db;
}

