#include "httplib.h"
#include <iostream>
#include <rapidjson/document.h>
#include <rapidjson/stringbuffer.h>
#include <rapidjson/writer.h>
#include "cpp_sdk.h"
#include <sys/time.h>
#include "cppspecdemo_db.h"

#define LogErr(FORMAT, ...) WWSPEC_LOG_ERR(FORMAT, ##__VA_ARGS__)
#define LogInfo(FORMAT, ...) WWSPEC_LOG_INFO(FORMAT, ##__VA_ARGS__)

using namespace std;
using namespace Tencent::WxWork;

std::string ToString(rapidjson::Value & value) {
    rapidjson::StringBuffer buffer;
    rapidjson::Writer<rapidjson::StringBuffer> writer(buffer);
    value.Accept(writer);
    return buffer.GetString();
}

class SpecSDKWrap: public SpecSDK {
public:
    using SpecSDK::SpecSDK;
    using SpecSDK::Invoke;
    int Invoke(const std::string &api_name, 
            const std::string &request,
            rapidjson::Document *resp, 
            string *response_data = nullptr) {
        string tmp;
        if (response_data == nullptr) {
            response_data = &tmp;
        }
        string errtype;
        int ret = _Invoke(api_name, request, *resp, *response_data, errtype);
        if (ret == 0) {
            LogInfo("INFO: api_name=%s,request=%s,response=%s",
                    api_name.c_str(), request.c_str(), response_data->c_str());
        } else {
            LogErr("ERROR: %s: api_name=%s,ret=%d,request=%s,response=%s",
                    errtype.c_str(), api_name.c_str(), ret, request.c_str(), response_data->c_str());
        }
        return ret;
    }
private:
    int _Invoke(const std::string &api_name, 
            const std::string &request,
            rapidjson::Document &resp, 
            std::string &response_data,
            std::string &errtype) {
        int ret = Invoke(api_name, request, &response_data);
        if (ret != 0) {
            errtype = "invoke fail";
            return ret;
        }
        resp.Parse(response_data);
        if (resp.HasParseError() || !resp.IsObject()) {
            errtype = "response not valid json";
            return -1; 
        }
        if (!resp.HasMember("errcode")) {
            errtype = "missing errcode";
            return -1; //企业微信接口异常
        }
        if (!resp["errcode"].IsInt()) {
            errtype = "errcode invalid";
            return -1; //企业微信接口异常
        }
        ret = resp["errcode"].GetInt();
        if (ret != 0) {
            errtype = "invoke_fail";
            return ret;
        }
        return 0;
    }
};

std::string GetJsonStringValue(
        rapidjson::Value & doc,
        const string &name,
        const string &def= "") {
    size_t pos;
    if ((pos = name.find('.')) == string::npos) {
        if (!doc.HasMember(name) || !doc[name].IsString()) {
            return def;
        }
        return doc[name].GetString();
    } else {
        string first = name.substr(0, pos);
        string second = name.substr(pos + 1);
        if (doc.HasMember(first) && doc[first].IsObject()) {
            return GetJsonStringValue(doc[first], second, def);
        } else {
            return def;
        }
    }
}

uint32_t GetJsonUint32Value(
        rapidjson::Value & doc,
        const string &name, 
        uint32_t def = 0) {
    size_t pos;
    if ((pos = name.find('.')) == string::npos) {
        if (!doc.HasMember(name) || !doc[name].IsUint()) {
            return def;
        }
        return doc[name].GetUint();
    } else {
        string first = name.substr(0, pos);
        string second = name.substr(pos + 1);
        if (doc.HasMember(first) && doc[first].IsObject()) {
            return GetJsonUint32Value(doc[first], second, def);
        } else {
            return def;
        }
    }
}


static string Wrap(const string &content, const std::string &name) {
    rapidjson::StringBuffer buffer;
    rapidjson::Writer<rapidjson::StringBuffer> writer(buffer);
    writer.StartObject();
    writer.Key(name);
    writer.String(content);
    writer.EndObject();
    return buffer.GetString();
}

static string GenErrorResponse(int errcode, const string &errmsg) {
    rapidjson::StringBuffer buffer;
    rapidjson::Writer<rapidjson::StringBuffer> writer(buffer);
    writer.StartObject();
    writer.Key("errcode");
    writer.Int(errcode);
    writer.Key("errmsg");
    writer.String(errmsg);
    writer.EndObject();
    return buffer.GetString();
}

static inline string GetFileNamePure(const std::string &path) {
    size_t pos = path.find_last_of('/');
    return pos == string::npos ? path : path.substr(pos + 1);
}

#define RETURN_ERROR(MSG) \
    do {\
        std::stringstream ss;\
        ss << GetFileNamePure(__FILE__) << ":" \
            << __func__ << ":" \
            << __LINE__  << ":"\
            << " " << (MSG);\
        return GenErrorResponse(710660, ss.str());\
    } while (0)

std::string HandleCallback(const SpecCallbackSDK & callback_sdk) {
    LogInfo("begin request: "
            "corpid=%s,"
            "agentid=%u,"
            "call_type=%u,"
            "is_async=%d,"
            "job_info=%s,"
            "ability_id=%s,"
            "notify_id=%s,"
            "data=%s",
            callback_sdk.GetCorpId().c_str(),
            callback_sdk.GetAgentId(),
            callback_sdk.GetCallType(),
            callback_sdk.GetIsAsync(),
            callback_sdk.GetJobInfo().c_str(),
            callback_sdk.GetAbilityId().c_str(),
            callback_sdk.GetNotifyId().c_str(),
            callback_sdk.GetData().c_str());

    string corpid = callback_sdk.GetCorpId();
    uint32_t agentid = callback_sdk.GetAgentId();
    SpecSDKWrap invoke_sdk(callback_sdk);

    if (callback_sdk.GetCallType() == 1) {
        //应用主动调用API的request_data
        rapidjson::Document reqdata;
        if (callback_sdk.GetData().empty()) {
            reqdata.Parse("{}");
        } else {
            reqdata.Parse(callback_sdk.GetData());
        }
        if(reqdata.HasParseError() || !reqdata.IsObject()) {
            RETURN_ERROR("request_data not valid json");
        }

        auto &ability_id = callback_sdk.GetAbilityId();
        //不同的ability_id对应不同的输入输出协议
        if (ability_id == "transfer") {
            //透传sdk的请求
            string api_name = GetJsonStringValue(reqdata, "api_name");
            string invoke_req = GetJsonStringValue(reqdata, "request");
            LogInfo("debug api_name[%s] request_data[%s]",
                    api_name.c_str(), invoke_req.c_str());
            string invoke_resp; 
            int ret = invoke_sdk.Invoke(api_name, invoke_req, &invoke_resp);
            if (ret != 0) {
                LogErr("invoke fail: ret=%d,api_name=%s", ret, api_name.c_str());
                RETURN_ERROR("invoke fail");
            }
            LogInfo("invoke succ: api_name=%s,req=%s,resp=%s",
                    api_name.c_str(), invoke_req.c_str(), invoke_resp.c_str());
            return Wrap(invoke_resp, "response");
        } else if (ability_id == "get_callback_data") {
            string notify_id = callback_sdk.GetNotifyId();
            string event_data;
            int ret = DB::Instance()->GetCallbackEvent(corpid, agentid, notify_id, &event_data);
            if (ret == 0) {
                return event_data;
            } else if (ret == 1) {
                RETURN_ERROR("invalid notify_id");
            } else {
                RETURN_ERROR("read notify_id fail");
            }
        } else if (ability_id.substr(0, 7) == "invoke_") {
            string api_name = ability_id.substr(7);
            string invoke_req = callback_sdk.GetData();
            LogInfo("debug ability_id[%s] api_name[%s] request_data[%s]",
                    ability_id.c_str(), api_name.c_str(), invoke_req.c_str());
            string invoke_resp; 
            int ret = invoke_sdk.Invoke(api_name, invoke_req, &invoke_resp);
            if (ret != 0) {
                LogErr("invoke fail: ret=%d,api_name=%s", ret, api_name.c_str());
                RETURN_ERROR("invalid fail");
            }
            LogInfo("invoke succ: api_name=%s,req=%s,resp=%s",
                    api_name.c_str(), invoke_req.c_str(), invoke_resp.c_str());
            return invoke_resp;
        }
        RETURN_ERROR("unknown ability_id");
    } else {
        //企业微信回调事件,暂存起来,通知应用主动来获取
        //产生"program_notify"事件
        //https://developer.work.weixin.qq.com/document/path/99843
        string raw_response;
        rapidjson::Document notify_resp;
        int ret = invoke_sdk.Invoke(
                "spec_notify_app", "{}", &notify_resp, &raw_response);
        LogInfo("invoke done: ret=%d", ret);
        string notify_id;
        if (ret == 0) {
            notify_id = GetJsonStringValue(notify_resp, "notify_id");
        }
        if (ret != 0 || notify_id.empty()) {
            LogErr("invalid resp: ret=%d", ret);
            RETURN_ERROR("call spec_notify_app fail");
        }
        ret = DB::Instance()->SetCallbackEvent(
                corpid, agentid, notify_id, callback_sdk.GetData());
        if (ret == 0) {
            return GenErrorResponse(0, "success");
        } else {
            RETURN_ERROR("save_callback_data_fail");
        }
    }
}

int main(int argc, char *argv[]) {
    if (argc >= 3) {
        LogInfo("start debug");
        Tencent::WxWork::SpecOpenDebugMode(argv[1], argv[2]);
    } else {
        LogInfo("start normal");
    }
    httplib::Server svr;
    svr.Post("/", [] (const httplib::Request & http_request, 
                httplib::Response & http_response) {
        map<string, string> request_headers;
        for (auto &item: http_request.headers) {
            LogInfo("header: %s: %s", item.first.c_str(), item.second.c_str());
            request_headers.emplace(item);
        }
        LogInfo("body: %s", http_request.body.c_str());
        SpecCallbackSDK callback_sdk(
                http_request.method, request_headers, http_request.body);
        string response;
        if (callback_sdk.IsOk()) {
            response = HandleCallback(callback_sdk);
        } 
        map<string, string> response_headers;
        string response_body;
        callback_sdk.BuildResponseHeaderBody(
                response, &response_headers, &response_body);
        string sContentType;
        for (auto &header: response_headers) {
            http_response.set_header(header.first, header.second);
            if(header.first == "Content-Type") {
                sContentType = header.second;
            }
        }
        if (sContentType.empty()) {
            sContentType = "text/plain";
        }
        http_response.set_content(response_body, sContentType);
    });

    LogInfo("listen start");
    if (!svr.listen("0.0.0.0", 8080)) {
        LogErr("listen fail");
    }
    return 0;
}

